/**
* FLV model class of the players MCV pattern.
* Handles playback of FLV files, HTTP streams and RTMP streams.
*
* @author	Jeroen Wijering
* @version	1.13
**/

/*
 * 
 * currentState : 0,1,2,3
 *  0 : => en pause => play:true,  pause=false (on peut démarrer)
 *  1 : => start    => play:false, pause:true  (on peut mettre en pause)
 *  2 : => on avance
 *  3 : => arrive en fin de vidéo 
 */

import com.jeroenwijering.players.*;


class com.jeroenwijering.players.FLVModel extends AbstractModel {


	/** array with extensions used by this model **/
	private var mediatypes:Array = new Array(
		"flv","rtmp","mp4","m4v","m4a","mov","3gp","3g2");
	/** NetConnection object reference **/
	private var connectObject:NetConnection;
	/** NetStream object reference **/
	private var streamObject:NetStream;
	/** Sound object reference **/
	private var soundObject:Sound;
	/** interval ID of the buffer update function **/
	private var loadedInterval:Number;
	/** current percentage of the video that's loaded **/
	private var currentLoaded:Number = 0;
	/** interval ID of the position update function **/
	private var positionInterval:Number;
	/** Duration metadata of the current video **/
	private var metaDuration:Number = 0;
	/** current state of the video that is playing **/
	private var currentState:Number;
	/** Current volume **/
	private var currentVolume:Number;
	/** MovieClip with "display" video Object  **/
	private var videoClip:MovieClip;
	/** object with keyframe times and positions, saved for PHP streaming **/
	private var metaKeyframes:Object = new Object();
	/** Boolean to check whether a stop event is fired **/
	private var stopFired:Boolean = false;
	/** Switch for FLV type currently played **/
	private var flvType:String;
	/** reference to the captions object for parsing captionate data **/
	public var capView:Object;

	/** PHP streaming mode */
	private var isPhpStreaming:Boolean = false;
	
	/** Constructor **/
	function FLVModel(vws:Array,ctr:AbstractController,cfg:Object,
		fed:Object,fcl:MovieClip) {
		super(vws,ctr,cfg,fed);
		connectObject = new NetConnection();
		videoClip = fcl;
		if(config["smoothing"] == "false") { 
			videoClip.display.smoothing = false;
			videoClip.display.deblocking = 1;
		} else {
			videoClip.display.smoothing = true;
			videoClip.display.deblocking = 4;
		}
		videoClip.createEmptyMovieClip("snd",videoClip.getNextHighestDepth());
		soundObject = new Sound(videoClip.snd);
	};


	/** Check which FLV type we use **/
	private function setItem(idx:Number) {
		super.setItem(idx);
		if(isActive == true) {
			isPhpStreaming = false;
			if(config["streamscript"] != undefined) { 
				flvType = "HTTP";
				if(config["streamscript"] != "lighttpd") {
					// PhilFlash : PHP Streaming mode
					isPhpStreaming = true;
				}
			} else if(feeder.feed[currentItem]["type"] == "rtmp") {
				flvType = "RTMP"; 
			} else { 
				flvType = "FLV"; 
			}
		}
	};


	/** Start a specific video **/
	private function setStart(pos:Number) {
		//trace("setStart pos="+pos+" currentURL="+currentURL+" =>"+(feeder.feed[currentItem]["file"])+" ,skipInStream="+skipInStream);
		if (skipInStream) {
			//trace("setStart REFUSED by skipInStream");
			return;
		}
		if (pos != undefined) { currentPosition = pos; }
		if(pos < 1) { 
			pos = 0; 
		} else if (pos > metaDuration - 1) { 
			pos = metaDuration - 1; 
		}
		// Start after a seekStop (or after loading)
		if (isPhpStreaming && feeder.feed[currentItem]["file"] == currentURL && seekStopPosition >= 0) {
			//trace(" START seekStopPosition ="+seekStopPosition+",pos="+pos+", streamObject="+streamObject+" ,waitEndSeek="+waitEndSeek);
			if (streamObject != null) {
				// close the stream 
				streamObject.close();
				videoClip.display.attachVideo(streamObject);
			}
			else {
				createPhpStreamObject(connectObject, false);
			}
			// Cases
			if (pos != undefined && pos > 0) {
				// Saut normal
				currentPosition = pos;	
			}
			else if (pos == undefined && seekStopPosition > 0) {
				// on est en arrêt, on saute et on redémarre 
				pos = seekStopPosition;
				currentPosition = pos;	
			}
			else if (pos == undefined && seekStopPosition == 0) {
				// on est en arrêt => on démarre 
				currentPosition = 0;
				pos = undefined;
				// on laisse pos = undefined pour ne pas excuter code qui suit
				streamObject.play(config["streamscript"] + "?file=" + currentURL);
			}
			seekStopPosition = -1;
			//trace(" START seek SET POS="+pos+",currentPosition ="+pos);
		}
		else if (isPhpStreaming && feeder.feed[currentItem]["file"] == currentURL && pos == 0) {
			// Saut en début
			currentPosition = 0;
			// On force 0.1 pour avoir le cas playKeyframe(currentPosition);
			pos = 0.1; 
		}
		else if (flvType=="RTMP" && feeder.feed[currentItem]["id"] != currentURL) {
			connectObject.connect(feeder.feed[currentItem]["file"]);
			currentURL = feeder.feed[currentItem]["id"];
			setStreamObject(connectObject);
			streamObject.play(currentURL);
		} else if(flvType != "RTMP" && 
			feeder.feed[currentItem]["file"] != currentURL) {
			connectObject.connect(null);
			currentURL = feeder.feed[currentItem]["file"];
			if(flvType == "HTTP" ) {
				setStreamObject(connectObject);
				if(config["streamscript"] == "lighttpd") {
					streamObject.play(currentURL);
				} else {
					streamObject.play(config["streamscript"] + 
					"?file=" + currentURL);
				}
			} else {
				setStreamObject(connectObject);
				streamObject.play(currentURL);
			}
		} else {
			if(pos != undefined) { streamObject.seek(pos); }
			streamObject.pause(false);
		}
		videoClip._visible = true;
		videoClip._parent.thumb._visible = false;
		if(flvType == "HTTP" && pos > 0) {
			playKeyframe(currentPosition);
		} else if (flvType == "FLV" && pos > 0) {
			streamObject.seek(currentPosition);
		} else if (flvType == "RTMP" && pos > 0) {
			streamObject.seek(currentPosition);
		}
		clearInterval(positionInterval);
		positionInterval = setInterval(this,"updatePosition",200);
		clearInterval(loadedInterval);
		loadedInterval = setInterval(this,"updateLoaded",200);
	};


	/** Read and broadcast the amount of the flv that's currently loaded **/
	private var updateLoadedCount = 0;
	private function updateLoaded() {
		updateLoadedCount++;
		if(flvType == "FLV") {
			var pct:Number = Math.round(streamObject.bytesLoaded/
				streamObject.bytesTotal*100);
		} else {
			var pct:Number = Math.round(streamObject.bufferLength/
				streamObject.bufferTime*100);
		}
		if(isNaN(pct)) { 
			updateLoadedCount = 0; // PhilFlash
			currentLoaded = 0;
			sendUpdate("load",0);
		} else if (pct > 95) {
			clearInterval(loadedInterval);
			currentLoaded = 100;
			// PhilFlash
			updateLoadedCount = 0;
			skipInStream = false;
			sendUpdate("load",100);
		} else if (pct != currentLoaded) { 
			currentLoaded= pct;
			sendUpdate("load",currentLoaded);
		}
		// Bug in PHP Streaming in special case => authorize operation with skipInStream = false
		if (skipInStream && updateLoadedCount >= 15) {
			//trace(" !!!! updateLoaded SET skipInStream = false");
			skipInStream = false;
			updateLoadedCount = 0;
		}
	};


	/** Read and broadcast the current position of the song **/
	var skipInStream:Boolean = false;
	private function updatePosition() {
		var pos = streamObject.time;
		// PhilFlash --> Wait the end of the skip
		if (skipInStream) {
			if (currentPosition == pos || Math.abs(currentPosition - pos) < 0.0001)	{
				skipInStream = false;			
			}
			else {
				// wait
				return;
			}
		}
		// <--
		if (config["bufferlength"] < 3) { 
			var mbl = 2;
		} else { 
			var mbl = config["bufferlength"] -1;
		}
		if(pos == currentPosition && currentState != 1 &&
			streamObject.bufferLength < mbl) {
			currentState = 1;
			sendUpdate("state",1);
		} else if (pos != currentPosition && currentState != 2) { 
			currentState = 2;
			sendUpdate("state",2);
		}
		// Special case for the end of the video for PHP Streaming
		if ((pos > (metaDuration-0.5)) && metaDuration != 0) {
			// The video is terminated
			//trace("PHP STREAMING video terminated");
			currentState = 3;
			videoClip._visible = false;
			videoClip._parent.thumb._visible = true;
			sendUpdate("state",3);
			sendCompleteEvent();
			stopFired = false;	
			return;
		}
		if (pos != currentPosition) {
			currentPosition = pos;
			if(metaDuration<currentPosition) {
				metaDuration = currentPosition;
			}
			sendUpdate("time",currentPosition,metaDuration-currentPosition);
		} else if (streamObject.bufferLength < mbl && stopFired == true) {
			currentState = 3;
			videoClip._visible = false;
			videoClip._parent.thumb._visible = true;
			sendUpdate("state",3);
			sendCompleteEvent();
			stopFired = false;
		}
	};


	/** Pause the video that's currently playing. **/
	private function setPause(pos:Number) {
		//trace("setPause pos =" +pos);
		if (skipInStream) {
			//trace("setPause REFUSED by skipInStream");
			return;
		}
		if(pos < 1) { pos = 0; }
		clearInterval(positionInterval);
		// PHPStreaming = skip when pause
		if (pos != undefined && isPhpStreaming) {
			//trace(" isPhpStreaming SET PAUSE pos="+pos+" ,currentPosition="+currentPosition); 
			if (streamObject != null) {
				streamObject.close();
				clearInterval(loadedInterval);
				streamObject = null;
			}
			// Special case for php streaming
			doSeekStop(pos);
			return;
		}
		if(pos != undefined) {
			currentPosition = pos;
			sendUpdate("time",currentPosition,
				Math.abs(metaDuration-currentPosition));
			streamObject.seek(currentPosition);
		}
		streamObject.pause(true);
		currentState = 0;
		sendUpdate("state",0);
	};


	/** Stop video and clear data. **/
	private function setStop(pos:Number) {
		//trace("setStop");
		clearInterval(loadedInterval);
		clearInterval(positionInterval);
		videoClip._visible = false;
		delete currentURL;
		delete currentLoaded;
		delete currentPosition;
		delete metaKeyframes;
		metaDuration = 0;
		currentLoaded = 0;
		stopFired = false;
		streamObject.close();
		delete streamObject;
		// for PHP Streaming
		clearInterval( seekStatusID ); 
		if (streamSeek != null) {
			streamSeek.close();
			delete streamSeek;
			streamSeek = null;
		}
		//
		clearInterval( loadStatusID ); 
		if (streamLoad != null) {
			streamLoad.close();
			delete streamLoad;
			streamLoad = null;
		}		
		skipInStream = false;
	};


	/** Set volume of the sound object. **/
	private function setVolume(vol:Number) {
		super.setVolume(vol);
		currentVolume = vol;
		soundObject.setVolume(vol);
	};


	/** Connect a new stream object to video/audio/callbacks **/
	private function setStreamObject(cnt:NetConnection) {
		var ref = this;
		currentLoaded = 0;
		sendUpdate("load",0);
		streamObject = new NetStream(cnt);
		streamObject.setBufferTime(config["bufferlength"]);
		streamObject.onMetaData = function(obj) {
			ref.sendUpdate("datarate",obj.videodatarate,obj.audiodatarate);
			obj.duration>1 ? ref.metaDuration = obj.duration: null;
			if(obj.width > 10) {
				ref.sendUpdate("size",obj.width,obj.height);
			}
			ref.frameRate = obj.framerate;
			ref.metaKeyframes = obj.keyframes;
			ref.prepareMetaTimes(ref.metaKeyframes); // PhilFlash
			if(ref.feeder.feed[ref.currentItem]['start'] > 0) {
				if(ref.flvType == "HTTP") {
					ref.playKeyframe(ref.feeder.feed[ref.currentItem]['start']);
				} else if (ref.flvType == "RTMP") {
					ref.setStart(ref.feeder.feed[ref.currentItem]['start']);
				}
			}
			delete obj;
			delete this.onMetaData;
		};
		streamObject.onStatus = function(object) {
			trace(object.code);
			if(object.code == "NetStream.Buffer.Flush" || 
				object.code == "NetStream.Play.Stop") {
				ref.stopFired = true;
			} else if (object.code == "NetStream.Play.StreamNotFound") {
				ref.currentState = 3;
				ref.videoClip._visible = false;
				ref.sendUpdate("state",3);
				ref.sendCompleteEvent();
				ref.stopFired = false;
			} else if (object.code == "NetStream.Play.Start") {
				ref.stopFired = false;
			}
		};
		streamObject.onCaption = function(cap:Array) {
			ref.capView.onCaptionate(cap);
		};
		videoClip.display.attachVideo(streamObject);
		videoClip.snd.attachAudio(streamObject);
	};

	// PhilFlash createPhpStreamObject
	// en fait, méthode setStreamObject mais avec observeMetaData en option
	//          si on a déjà observeMetaData, on ne recommence pas 
	private function createPhpStreamObject(cnt:NetConnection, observeMetaData:Boolean) {
		var ref = this;
		connectObject.connect(null);
		streamObject = new NetStream(cnt);
		streamObject.setBufferTime(config["bufferlength"]);
		if (observeMetaData) {
			streamObject.onMetaData = function(obj) {
				//trace(" streamObject onMetaData READ");
				ref.sendUpdate("datarate",obj.videodatarate,obj.audiodatarate);
				obj.duration>1 ? ref.metaDuration = obj.duration: null;
				if(obj.width > 10) {
					ref.sendUpdate("size",obj.width,obj.height); 
					// PhilFlash
					//ref.videoClip.display.deblocking = 4;
					//ref.videoClip.display.smoothing = true;
				}
				ref.frameRate = obj.framerate;
				ref.metaKeyframes = obj.keyframes;
				delete obj;
				delete this.onMetaData;
			};		
		}
		streamObject.onStatus = function(object) {
			trace(object.code);
			if(object.code == "NetStream.Buffer.Flush" || 
				object.code == "NetStream.Play.Stop") {
				ref.stopFired = true;
			} else if (object.code == "NetStream.Play.StreamNotFound") {
				ref.currentState = 3;
				ref.videoClip._visible = false;
				ref.sendUpdate("state",3);
				ref.sendCompleteEvent();
				ref.stopFired = false;
			} else if (object.code == "NetStream.Play.Start") {
				ref.stopFired = false;
			}
		};
		streamObject.onCaption = function(cap:Array) {
			ref.capView.onCaptionate(cap);
		};
		videoClip.display.attachVideo(streamObject);
		videoClip.snd.attachAudio(streamObject);
	}


	/** Play from keyframe position from metadata **/
	/*
	private function playKeyframe(pos:Number) {
		for (var i=0; i< metaKeyframes.times.length; i++) {
			if((metaKeyframes.times[i] <= pos) && 
				(metaKeyframes.times[i+1] >= pos)) {
				if(config["streamscript"] == "lighttpd") {
					streamObject.play(currentURL+"?start="+
						metaKeyframes.filepositions[i]);
				} else {
					streamObject.play(config["streamscript"]+"?file="+
						currentURL+"&pos="+metaKeyframes.filepositions[i]);
				}
				break;
			}
		}
	};
	*/

	// For a bug in computeIndiceForPosition for special value like 1014.466
	private function prepareMetaTimes(meta)
	{
		// For the bug of value 1014.466 => 
		//  trace(" bug"+(1014.466-Number("1014.466")));  => 1.13686837721616e-13
		var len = meta.times.length;
		for (var i=0; i< len; i++) {
			var val:Number = meta.times[i];
			meta.times[i] = Number((""+val));
		}
	}

	/** Play from keyframe position from metadata **/
	/** Special case for PHP streaming **/
	private function playKeyframe(pos:Number) {	
		//trace(" PHP STREAM playKeyframe ="+pos+" streamObject="+streamObject);	
		var computeIndice = computeIndiceForPosition(pos);
		if(config["streamscript"] == "lighttpd") {
			streamObject.play(currentURL+"?start="+metaKeyframes.filepositions[computeIndice]);
		} else {
			streamObject.close();
			// update position of scrubbar
			currentPosition = metaKeyframes.times[computeIndice];
			metaDuration < currentPosition ? metaDuration = currentPosition: null;
			sendUpdate("time",currentPosition,metaDuration-currentPosition);
			//		
			skipInStream = true;
			streamObject.play(config["streamscript"]+"?file="+
				currentURL+"&pos="+metaKeyframes.filepositions[computeIndice]);	
		}
	}	

	// ----- PHP Streaming : Special case for autostart = false  -----
	// --> read metadata, no sound, buffertime very small to load the first image of the video
	private var streamLoad:NetStream;
	private var loadStatusID:Number;
	private var loadConnectObject:NetConnection = null;
	
	public function doLoad(pos:Number):Void {
		seekStopPosition = -1;
		sendUpdate("load",100);
		var ref = this;
		if (loadConnectObject == null) {
			loadConnectObject = new NetConnection();
			loadConnectObject.connect(null);
		}
		streamLoad = new NetStream(loadConnectObject);
		streamLoad.setBufferTime(0.1);

		 // -- read metadata --
		streamLoad.onMetaData = function(obj) {
			ref.sendUpdate("datarate",obj.videodatarate,obj.audiodatarate);
			obj.duration>1 ? ref.metaDuration = obj.duration: null;
			if(obj.width > 10) {
				ref.sendUpdate("size",obj.width,obj.height); 
			}
			ref.frameRate = obj.framerate;
			ref.metaKeyframes = obj.keyframes;
			ref.prepareMetaTimes(ref.metaKeyframes); // PhilFlash
			delete obj;
			delete this.onMetaData;
		};
		// no sound 
		videoClip.display.attachVideo(streamLoad);
		// -- 
		currentURL = feeder.feed[currentItem]["file"];
		loadStatusID = setInterval(this, "videoLoadStatus", 100);
		streamLoad.play(config["streamscript"]+"?file="+currentURL);
		streamLoad.pause(true);
		streamLoad.seek(0);
		// --
		currentState = 0;
		sendUpdate("state",0);
	}
	
	private function videoLoadStatus():Void {
		var bloaded = streamLoad.bufferLength;
		if (bloaded> 0.1) {
			clearInterval( loadStatusID ); 	
			streamLoad.close();
			delete streamLoad;
			streamLoad = null;
			sendUpdate("time",0,metaDuration);
			seekStopPosition = 0;
		}
	} 
	
	// ----- PHP Streaming : Special case for seek (when the video is in pause mode)  -----
	private var streamSeek:NetStream;	
	private var seekStatusID:Number;
	private var seekStopPosition:Number; // la position pour le prochain start
	private var waitEndSeek:Number = -1; // pour savoir si on effectue un seekStop
	
	public function doSeekStop(pos:Number):Void {
		var ref = this;
		//trace(" DO SEEK STOP VIDEO waitEndSeek="+waitEndSeek+" pos="+pos);
		if (loadConnectObject == null) {
			loadConnectObject = new NetConnection();
			loadConnectObject.connect(null);
		}
		
		// already seeking
		if (waitEndSeek >= 0) return;
		
		if (streamSeek != null) {
			// on est dèjà en seek => a priori n'arrive pas mais on ne sait jamais 
			streamSeek.close();
			clearInterval( seekStatusID ); 	
		} 
		
		videoClip._visible = true;
		
		streamSeek = new NetStream(loadConnectObject);
		streamSeek.setBufferTime(0.1);

		videoClip.display.attachVideo(streamSeek);
		// -- 
		var newIndice = computeIndiceForPosition(pos);
		seekStopPosition = metaKeyframes.times[newIndice];
		var newPosition = metaKeyframes.filepositions[newIndice];
		//
		clearInterval( seekStatusID ); 	
		seekStatusID = setInterval(this, "videoSeekStatus", 100);
		if (pos == 0)
		{
			streamSeek.play(config["streamscript"]+"?file="+currentURL);
		}
		else {
			streamSeek.play(config["streamscript"]+"?file="+currentURL+"&pos="+newPosition);
		}
		streamSeek.pause(true);
		streamSeek.seek(0);
		// 
		currentPosition = pos;
		waitEndSeek = pos; // pour repositionner la scrubbar correctement pour les sauts fous
		sendUpdate("time",currentPosition,metaDuration-currentPosition);
		currentState = 0;
		sendUpdate("state",0);
	}
	
	private function videoSeekStatus():Void {
		//trace(" videoSeekStatus loaded/total="+streamSeek.bytesLoaded+"/"+streamSeek.bytesTotal+" bufferLength="+streamSeek.bufferLength+" , duration="+metaDuration);
		var bloaded = streamSeek.bufferLength;
		if (bloaded> 0.1) {
			//trace(" videoSeekStatus STOP");
			clearInterval( seekStatusID ); 	
			streamSeek.close();
			delete streamSeek;
			streamSeek = null;
			// on remets à jour scrubbar (pour les sauts fous en avant et arrière)
			currentPosition = waitEndSeek;
			sendUpdate("time",currentPosition,metaDuration-currentPosition);
			waitEndSeek = -1;
		}
	} 
	

	// ----- Fasted method to compute the start position in the video  -----	
	private function computeIndiceForPosition(newPosition:Number):Number {	
		var len:Number = metaKeyframes.times.length;
		var computeIndice:Number = 0;
		
		if (newPosition <= metaKeyframes.times[0]) {
			// seek before 0 => seek to first position
			computeIndice = 0;
		}
		else if (newPosition >= metaKeyframes.times[len-1]) {
			// seek at end or superior => back to 0
			computeIndice = 0;
		}
		else {
			var linearSearchTolerance:Number = 40;
			var startIndex:Number = 0;
			var endIndex:Number = len;
			var newStart:Number = 0;
			var newEnd:Number = 0;

			while ((endIndex - startIndex) > linearSearchTolerance) {
				var newMax = endIndex - startIndex;
				var k:Number = (newMax>>1);  // div 2
				k = startIndex + k;
				var timeMiddle:Number = metaKeyframes.times[k];
				//
				newStart = startIndex;
				newEnd = k;
				if (newPosition >= timeMiddle) { newStart = k; newEnd = endIndex; }
				startIndex = newStart;
				endIndex = newEnd;
			}
			// Search 
			for (var i:Number = startIndex; i < endIndex; i++) {
				if((metaKeyframes.times[i] <= newPosition) && (metaKeyframes.times[i+1] > newPosition)) {
					computeIndice = i;
					break;
				}
			}
		}
		return computeIndice;
	}


}